/*
    Project: TCGINA
    Version: V1.25
    File   : tc.c

    Copyright (c) 2005-2008 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tcpriv.h"
#include "secmem.h"
#include <tchar.h>
#include <stdio.h>

#define SUPPORT_AUTO_HIDDEN_VOLUME_PROTECTION

typedef struct
{
    VOID  (WINAPI *Close)(PVOID);
    BOOL  (WINAPI *Mount)(PTC_CONTEXT, LPCTSTR pszTCFileName,
            LPCSTR paszPassword, LPCTSTR pszDrive, DWORD, LPCTSTR pszKeyFiles,
            BOOL   fProtectHiddenVolume,
            LPCSTR paszHiddenVolumePassword, LPCTSTR pszHiddenVolumeKeyFiles);
    BOOL  (WINAPI *WipeCache)(PTC_CONTEXT);
    BOOL  (WINAPI *Dismount)(PTC_CONTEXT, LPCTSTR pszDrive, DWORD dwFlags);
    BOOL  (WINAPI *GetVolumeFileName)(PTC_CONTEXT, LPCTSTR pszDrive,
            LPTSTR pszDest, DWORD nSize);
    BOOL  (WINAPI *GetMountedDrives)(PTC_CONTEXT, PDWORD pdwDrives, DWORD dwFlags);
    VOID  (WINAPI *RefreshDriveLetters)(PTC_CONTEXT);
} TC_FUNCTION_SET, *PTC_FUNCTION_SET;

#define TC_DEFINE_FUNCTION_SET(x) \
PVOID WINAPI TC##x##_Open(VOID); \
VOID  WINAPI TC##x##_Close(PVOID); \
BOOL  WINAPI TC##x##_Mount(PTC_CONTEXT, LPCTSTR, LPCSTR, LPCTSTR, DWORD, LPCTSTR, \
  BOOL, LPCSTR, LPCTSTR); \
BOOL  WINAPI TC##x##_WipeCache(PTC_CONTEXT); \
BOOL  WINAPI TC##x##_Dismount(PTC_CONTEXT, LPCTSTR, DWORD); \
BOOL  WINAPI TC##x##_GetVolumeFileName(PTC_CONTEXT, LPCTSTR, LPTSTR, DWORD); \
BOOL  WINAPI TC##x##_GetMountedDrives(PTC_CONTEXT, PDWORD, DWORD); \
VOID  WINAPI TC##x##_RefreshDriveLetters(PTC_CONTEXT); \
TC_FUNCTION_SET tcfs##x = { \
    TC##x##_Close, \
    TC##x##_Mount, \
    TC##x##_WipeCache, \
    TC##x##_Dismount, \
    TC##x##_GetVolumeFileName, \
    TC##x##_GetMountedDrives, \
    TC##x##_RefreshDriveLetters \
}

#ifdef IMPLEMENT_TC_EXE_INTERFACE
    TC_DEFINE_FUNCTION_SET(EXE);
#endif
#ifdef IMPLEMENT_TC_SYS_INTERFACE
    TC_DEFINE_FUNCTION_SET(SYS);
#endif

BOOL WINAPI TC_VolumeHostExists(LPCTSTR pszHostName)
{
    TC_PARTITION_DATA pd;

    if (TC_GetPartitionData(&pd, pszHostName)) return TRUE;

    return TC_FileExists(pszHostName);
}

TC_HANDLE WINAPI TC_Open(DWORD dwFlags) {

    PTC_CONTEXT pContext;

    TC_Assert((dwFlags & 0x8000ffff) == 0);

    pContext = (PTC_CONTEXT) TC_Calloc(sizeof(TC_CONTEXT));
    if (pContext == NULL) {
        TC_DbgPutsA("Out of memory in TC_Open()!\n");
        return NULL;
    }

    if (dwFlags == 0) dwFlags = TC_ALL_INTERFACES;

#ifdef IMPLEMENT_TC_SYS_INTERFACE
    if (dwFlags & TC_SYS_INTERFACE) {
        /* Try to use the device driver directly */
        pContext->pContext = TCSYS_Open();
        if (pContext->pContext) {
            pContext->pFunctionSet = &tcfsSYS;
            return pContext;
        }
    }
#endif

#ifdef IMPLEMENT_TC_EXE_INTERFACE
    if (dwFlags & TC_EXE_INTERFACE) {
        /* Try to use TrueCrypt.exe */
        pContext->pContext = TCEXE_Open();
        if (pContext->pContext) {
            pContext->pFunctionSet = &tcfsEXE;
            return pContext;
        }
    }
#endif

    TC_Free(pContext);

    return NULL;
}

#define TC_FUNCTION_SET(pCtx) \
  (*(PTC_FUNCTION_SET)((PTC_CONTEXT)pCtx)->pFunctionSet)
#define TC_PCONTEXT(pCtx) ((PTC_CONTEXT)pCtx)

void WINAPI TC_Close(TC_HANDLE hTC)
{
    TC_Assert(hTC);

    if (hTC) {
        TC_FUNCTION_SET(hTC).Close(TC_PCONTEXT(hTC)->pContext);
        TC_Free(hTC);
    }
}

BOOL WINAPI TC_MountWithAnsiPassword(TC_HANDLE hTC, LPCTSTR pszTCFileName,
  LPCSTR paszPassword, LPCTSTR pszDrive, DWORD dwFlags,
  LPCTSTR pszKeyFiles, BOOL fProtectHiddenVolume,
  LPCSTR paszHiddenVolumePassword, LPCTSTR pszHiddenVolumeKeyFiles)
{
    LPSTR pTemp;
    BOOL  fResult;
    TC_Assert(hTC);
    TC_Assert(pszTCFileName);
    TC_Assert(pszDrive);
    TC_Assert((dwFlags & 0xffff0000) == 0);

    if (hTC == NULL) {
        TC_SetLastError(hTC, ERROR_TC_INVALID_HANDLE);
        return FALSE;
    }

#ifndef SUPPORT_AUTO_HIDDEN_VOLUME_PROTECTION
    return TC_FUNCTION_SET(hTC).Mount(TC_PCONTEXT(hTC), pszTCFileName,
      paszPassword, pszDrive, dwFlags, pszKeyFiles, fProtectHiddenVolume,
      paszHiddenVolumePassword, pszHiddenVolumeKeyFiles);
#else
    if (TC_FUNCTION_SET(hTC).Mount(TC_PCONTEXT(hTC), pszTCFileName,
      paszPassword, pszDrive, dwFlags, pszKeyFiles, fProtectHiddenVolume,
      paszHiddenVolumePassword, pszHiddenVolumeKeyFiles)) {
        return TRUE;
    }
    if (fProtectHiddenVolume || paszPassword == NULL) {
        return FALSE;
    }

    /* Try to split password */
    pTemp = (LPSTR)strchr(paszPassword, ' ');
    if (pTemp == NULL) return FALSE;

    /* Try to mount with hidden volume protection */
    *pTemp = '\0';
    fResult = TC_FUNCTION_SET(hTC).Mount(TC_PCONTEXT(hTC), pszTCFileName,
      paszPassword, pszDrive, dwFlags, pszKeyFiles, TRUE, pTemp + 1,
      pszKeyFiles);
    *pTemp = ' ';

    return fResult;
#endif
}

BOOL WINAPI TC_Mount(TC_HANDLE hTC, LPCTSTR pszTCFileName,
  LPCTSTR pszPassword, LPCTSTR pszDrive, DWORD dwFlags,
  LPCTSTR pszKeyFiles, BOOL fProtectHiddenVolume,
  LPCTSTR pszHiddenVolumePassword, LPCTSTR pszHiddenVolumeKeyFiles)
{
#ifdef UNICODE
    BOOL  fResult;
    CHAR  *paszPassword;
    CHAR  *paszHiddenVolumePassword;
    DWORD dwErrorCode;
#endif

    TC_Assert(hTC);
    TC_Assert(pszTCFileName);
    TC_Assert(pszPassword);
    TC_Assert(pszDrive);
    TC_Assert((dwFlags & 0xffff0000) == 0);

#ifdef UNICODE
    if (pszPassword == NULL) {
        pszPassword = TEXT("");
    }
    if (pszHiddenVolumePassword == NULL) {
        pszHiddenVolumePassword = TEXT("");
    }
    paszPassword = sec_malloc(2*TC_MAX_PASSWORD);
    if (paszPassword == NULL) {
        TC_DbgPutsA("Out of secure memory in TC_Mount()!\n");
        TC_SetLastError(hTC, ERROR_TC_OUT_OF_SECURE_MEMORY);
        return FALSE;
    }
    if (WideCharToMultiByte(CP_ACP, 0, pszPassword, -1, paszPassword,
      TC_MAX_PASSWORD, NULL, NULL) == 0) {
        TC_DbgPutsA("Cannot convert password in TC_Mount()!\n");
        sec_free(paszPassword);
        TC_SetLastError(hTC, ERROR_TC_INVALID_PASSWORD);
        return FALSE;
    }
    paszHiddenVolumePassword = paszPassword + TC_MAX_PASSWORD;
    if (WideCharToMultiByte(CP_ACP, 0, pszHiddenVolumePassword, -1,
      paszHiddenVolumePassword, TC_MAX_PASSWORD, NULL, NULL) == 0) {
        TC_DbgPutsA("Cannot convert hidden volume password in TC_Mount()!\n");
        sec_free(paszPassword);
        TC_SetLastError(hTC, ERROR_TC_INVALID_PASSWORD);
        return FALSE;
    }
    fResult = TC_MountWithAnsiPassword(hTC, pszTCFileName,
      paszPassword, pszDrive, dwFlags, pszKeyFiles, fProtectHiddenVolume,
      paszHiddenVolumePassword, pszHiddenVolumeKeyFiles);
    dwErrorCode = TC_GetLastError(hTC);
    sec_free(paszPassword);
    TC_SetLastError(hTC, dwErrorCode);
    return fResult;
#else
    if (hTC == NULL) {
        TC_SetLastError(hTC, ERROR_TC_INVALID_HANDLE);
        return FALSE;
    }
    return TC_MountWithAnsiPassword(hTC, pszTCFileName,
      pszPassword, pszDrive, dwFlags, pszKeyFiles, fProtectHiddenVolume,
      pszHiddenVolumePassword, pszHiddenVolumeKeyFiles);
#endif
}

BOOL WINAPI TC_WipeCache(TC_HANDLE hTC) {

    TC_Assert(hTC);

    if (hTC == NULL) return FALSE;

    return TC_FUNCTION_SET(hTC).WipeCache(TC_PCONTEXT(hTC));
}

BOOL WINAPI TC_Dismount(TC_HANDLE hTC, LPCTSTR pszDrive, DWORD dwFlags)
{
    TC_Assert(hTC);
    TC_Assert((dwFlags & 0xffff0000) == 0);

    if (hTC == NULL) return FALSE;

    return TC_FUNCTION_SET(hTC).Dismount(TC_PCONTEXT(hTC), pszDrive, dwFlags);
}


BOOL WINAPI TC_GetVolumeFileName(TC_HANDLE hTC, LPCTSTR pszDrive,
  LPTSTR pszDest, DWORD nSize)
{
    TC_Assert(hTC);
    TC_Assert(pszDrive);
    TC_Assert(pszDest);

    if (hTC == NULL) return FALSE;

    return TC_FUNCTION_SET(hTC).GetVolumeFileName(TC_PCONTEXT(hTC), pszDrive,
      pszDest, nSize);
}

BOOL WINAPI TC_GetMountedDrives(TC_HANDLE hTC, PDWORD pdwDrives, DWORD dwFlags)
{
    TC_Assert(hTC);
    TC_Assert(pdwDrives);

    if (hTC == NULL) return FALSE;

    return TC_FUNCTION_SET(hTC).GetMountedDrives(TC_PCONTEXT(hTC), pdwDrives, dwFlags);
}

DWORD WINAPI TC_GetLastError(TC_HANDLE hTC)
{
    TC_Assert(hTC);

    if (hTC == NULL) return ERROR_TC_INVALID_HANDLE;

    return TC_PCONTEXT(hTC)->dwErrorCode;
}

BOOL WINAPI TC_SetLastError(TC_HANDLE hTC, DWORD dwErrorCode)
{
    TC_Assert(hTC);

    if (hTC == NULL) return FALSE;

    TC_PCONTEXT(hTC)->dwErrorCode = dwErrorCode;

    return TRUE;
}

BOOL WINAPI TC_GetErrorString(DWORD dwErrorCode, LPTSTR pszText,
  DWORD nSize)
{
    BOOL fResult;

    if ((dwErrorCode & 0x20000000) == 0) {

        LPTSTR pszBuffer = NULL;

        if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
          FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ARGUMENT_ARRAY,
          NULL, dwErrorCode, LANG_NEUTRAL, (LPTSTR) &pszBuffer, 0, NULL)) {
            fResult = TC_CopyString(pszText, nSize, pszBuffer);
            LocalFree(pszBuffer);
        } else {
            fResult = _sntprintf(pszText, nSize,
              TEXT("Win32 API Error Code = 0x%08x"), (unsigned)dwErrorCode) > 0;
        }

    } else if (dwErrorCode >= ERROR_TC_GENERIC &&
      dwErrorCode <= ERROR_TC_INVALID_PROFILE_IMAGE_PATH) {

        LPCTSTR rgszError[] = {
            TEXT("Unspecific error."),
            TEXT("Unknown error."),
            TEXT("Invalid pointer."),
            TEXT("Invalid handle."),
            TEXT("Invalid password."),
            TEXT("Invalid filename."),
            TEXT("Invalid drive letter."),
            TEXT("Drive letter not free."),
            TEXT("Partition is already mounted."),
            TEXT("File is already mounted."),
            TEXT("File not found."),
            TEXT("Out of secure memory."),
            TEXT("Cannot process keyfiles."),
            TEXT("Cannot find mounted TrueCrypt volume."),
            TEXT("Cannot create DOS device."),
            TEXT("Profile image path not found.")
        };

        fResult = TC_CopyString(pszText, nSize,
          rgszError[dwErrorCode-ERROR_TC_GENERIC]);
    } else {
        fResult = _sntprintf(pszText, nSize,
          TEXT("TCGINA Error Code = 0x%08x"), (unsigned)dwErrorCode) > 0;
    }

    return fResult;
}

VOID WINAPI TC_RefreshDriveLetters(TC_HANDLE hTC)
{
    TC_Assert(hTC);

    if (hTC == NULL) return;

    TC_FUNCTION_SET(hTC).RefreshDriveLetters(TC_PCONTEXT(hTC));
}